@@ -3,6 +3,7 @@ import os
 import random
 from datetime import datetime, timedelta
 from pathlib import Path
+from functools import partial
 from typing import Optional, Dict, List, Tuple, Union, Any
 import time
 
@@ -10,7 +11,7 @@ import numpy as np
 import torch
 
 from rtg import log, yaml
-from rtg.data.dataset import (TSVData, BatchIterable, LoopingIterable, SqliteFile)
+from rtg.data.dataset import (TSVData, BatchIterable, LoopingIterable, SqliteFile, GenerativeBatchIterable)
 from rtg.data.codec import Field, SPField, NLField, PretrainMatchField
 from rtg.utils import IO, line_count
 
@@ -32,7 +33,8 @@ class BaseExperiment:
         log.info(f"Initializing an experiment. Directory = {work_dir}")
         self.read_only = read_only
         self.work_dir = work_dir
-        self.log_file = work_dir / 'rtg.log'
+        self.log_dir = work_dir / 'logs'
+        self.log_file = self.log_dir / 'rtg.log'
         self.data_dir = work_dir / 'data'
         self.model_dir = work_dir / 'models'
         self._config_file = work_dir / 'conf.yml'
@@ -54,6 +56,7 @@ class BaseExperiment:
 
         self.train_file = self.data_dir / 'train.tsv.gz'
         self.train_db = self.data_dir / 'train.db'
+        self.train_db_tmp = self.data_dir / 'train.db.tmp'
         self.finetune_file = self.data_dir / 'finetune.db'
         self.valid_file = self.data_dir / 'valid.tsv.gz'
         self.combo_file = self.data_dir / 'combo.tsv.gz'
@@ -61,7 +64,7 @@ class BaseExperiment:
         self.samples_file = self.data_dir / 'samples.tsv.gz'
 
         if not read_only:
-            for _dir in [self.model_dir, self.data_dir]:
+            for _dir in [self.model_dir, self.data_dir, self.log_dir]:
                 if not _dir.exists():
                     _dir.mkdir(parents=True)
 
@@ -171,7 +174,9 @@ class BaseExperiment:
         :param desc: True to sort in reverse (default); False to sort in ascending
         :return: list of model paths
         """
-        paths = self.model_dir.glob('model_*.pkl')
+        paths = list(self.model_dir.glob('model_*.pkl'))
+        if not paths:
+            paths = list(self.model_dir.glob('embeddings_*.gz'))
         sorters = {
             'valid_score': self._path_to_validn_score,
             'total_score': self._path_to_total_score,
@@ -190,8 +195,10 @@ class BaseExperiment:
         :return: Tuple[Optional[Path], step_num:int]
         """
         models = self.list_models(sort=sort, desc=desc)
+        print(models)
         if models:
-            step, train_score, valid_score = models[0].name.replace('.pkl', '').split('_')[-3:]
+            name = models[0].name.replace('.pkl', '').replace('.txt.gz', '')
+            step, train_score, valid_score = name.split('_')[-3:]
             return models[0], int(step)
         else:
             return None, 0
@@ -272,7 +279,7 @@ class TranslationExperiment(BaseExperiment):
         assert not (self.shared_field and self.src_field)
         assert not (self.shared_field and self.tgt_field)
         # both are set or both are unset
-        assert (self.src_field is None) == (self.tgt_field is None)
+        #assert (self.src_field is None) == (self.tgt_field is None)
 
         self._unsupervised = self.model_type in {'binmt', 'rnnlm', 'tfmlm'}
         if self._unsupervised:
@@ -299,8 +306,13 @@ class TranslationExperiment(BaseExperiment):
 
     def pre_process_parallel(self, args: Dict[str, Any]):
         # check if files are parallel
-        self.check_line_count('training', args['train_src'], args['train_tgt'])
         self.check_line_count('validation', args['valid_src'], args['valid_tgt'])
+        if 'spark' in self.config:
+            log.warning(f"Spark backend detected: line count on training data is skipped")
+        else:
+            log.warning(f"Going to count lines. If this is a big dataset, it will take long time")
+            self.check_line_count('training', args['train_src'], args['train_tgt'])
+
         xt_args = dict(no_split_toks=args.get('no_split_toks'),
                        char_coverage=args.get('char_coverage', 0))
         if args.get('shared_vocab'):  # shared vocab
@@ -417,7 +429,8 @@ class TranslationExperiment(BaseExperiment):
                    self.tgt_vocab)
 
     def _pre_process_parallel(self, src_key: str, tgt_key: str, out_file: Path,
-                              args: Optional[Dict[str, Any]] = None, line_check=True):
+                              args: Optional[Dict[str, Any]] = None, line_check=True,
+                              split_ratio: float = 0.):
         """
         Pre process records of a parallel corpus
         :param args: all arguments for 'prep' task
@@ -435,12 +448,12 @@ class TranslationExperiment(BaseExperiment):
                 f'{args[src_key]} and {args[tgt_key]} must have same number of lines'
         # create Piece IDs
         s_time = time.time()
-        reader_func = TSVData.read_raw_parallel_recs_parallel if self.codec_supports_multiproc \
-            else TSVData.read_raw_parallel_recs
+        reader_func = TSVData.read_raw_parallel_recs
         parallel_recs = reader_func(
             args[src_key], args[tgt_key], args['truncate'], args['src_len'], args['tgt_len'],
-            src_tokenizer=self.src_vocab.encode_as_ids, tgt_tokenizer=self.tgt_vocab.encode_as_ids)
-        if out_file.name.endswith('.db'):
+            src_tokenizer=partial(self.src_vocab.encode_as_ids, split_ratio=split_ratio),
+            tgt_tokenizer=partial(self.tgt_vocab.encode_as_ids, split_ratio=split_ratio))
+        if any([out_file.name.endswith(suf) for suf in ('.db', '.db.tmp')]):
             SqliteFile.write(out_file, records=parallel_recs)
         else:
             TSVData.write_parallel_recs(parallel_recs, out_file)
@@ -554,6 +567,28 @@ class TranslationExperiment(BaseExperiment):
                 _write_dict(report, Path(str(outp) + '.report.txt'))
                 _write_emb_matrix(emb_matrix, str(outp))
 
+    def shrink_vocabs(self):
+        assert self.codec_name == 'nlcodec', 'Only nlcodec supports shrinking of vocabs'
+        args = self.config['prep']
+
+        if self.shared_vocab:
+            corpus = [args[key] for key in ['train_src', 'train_tgt', 'mono_src', 'mono_tgt']
+                      if args.get(key)]
+            remap_src = self.shared_vocab.shrink_vocab(files=corpus, min_freq=1,
+                                                    save_at=self._shared_field_file)
+            remap_tgt = remap_src
+        else:
+            corpus_src = [args[key] for key in ['train_src', 'mono_src'] if args.get(key)]
+            remap_src = self.src_vocab.shrink_vocab(files=corpus_src, min_freq=1,
+                                                     save_at=self._src_field_file)
+            corpus_tgt = [args[key] for key in ['train_tgt', 'mono_tgt'] if args.get(key)]
+            remap_tgt = self.tgt_vocab.shrink_vocab(files=corpus_tgt, min_freq=1,
+                                                     save_at=self._tgt_field_file)
+        self.reload_vocabs()
+        self.model_args['src_vocab'] = len(self.src_vocab)
+        self.model_args['tgt_vocab'] = len(self.tgt_vocab)
+        return remap_src, remap_tgt
+
     def inherit_parent(self):
         parent = self.config['parent']
         parent_exp = TranslationExperiment(parent['experiment'], read_only=True)
@@ -596,6 +631,38 @@ class TranslationExperiment(BaseExperiment):
             log.info(f"Saving parent model's state to {self.parent_model_state}")
             torch.save(avg_state, self.parent_model_state)
 
+        shrink_spec = parent.get('shrink')
+        if shrink_spec:
+            remap_src, remap_tgt = self.shrink_vocabs()
+            def map_rows(mapping: List[int], source: torch.Tensor, name=''):
+                assert max(mapping) < len(source)
+                target = torch.zeros((len(mapping), *source.shape[1:]),
+                                     dtype=source.dtype, device=source.device)
+                for new_idx, old_idx in enumerate(mapping):
+                    target[new_idx] = source[old_idx]
+                log.info(f"Mapped {name} {source.shape} --> {target.shape} ")
+                return target
+
+            """ src_embed.0.lut.weight [N x d]
+                tgt_embed.0.lut.weight [N x d]
+                generator.proj.weight [N x d]
+                generator.proj.bias [N] """
+            if remap_src:
+                key = 'src_embed.0.lut.weight'
+                avg_state[key] = map_rows(remap_src, avg_state[key], name=key)
+            if remap_tgt:
+                map_keys = ['tgt_embed.0.lut.weight', 'generator.proj.weight', 'generator.proj.bias']
+                for key in map_keys:
+                    if key not in avg_state:
+                        log.warning(f'{key} not found in avg_state of parent model. Mapping skipped')
+                        continue
+                    avg_state[key] = map_rows(remap_tgt, avg_state[key], name=key)
+            if self.parent_model_state.exists():
+                self.parent_model_state.rename(self.parent_model_state.with_suffix('.orig'))
+            torch.save(avg_state, self.parent_model_state)
+            self.persist_state()  # this will fix src_vocab and tgt_vocab of model_args conf
+
+
     def pre_process(self, args=None, force=False):
         if self.has_prepared() and not force:
             log.warning("Already prepared")
@@ -720,24 +787,41 @@ class TranslationExperiment(BaseExperiment):
                 [('src_len', 'max_src_len'), ('tgt_len', 'max_tgt_len'), ('truncate', 'truncate')]
                 if ik in prep_args}
 
-    def get_train_data(self, batch_size: int, steps: int = 0, sort_by='eq_len_rand_batch',
-                       batch_first=True, shuffle=False, fine_tune=False, keep_in_mem=False):
-        inp_file = self.train_db if self.train_db.exists() else self.train_file
+    def get_train_data(self, batch_size:  Union[int, Tuple[int,int]], steps: int = 0, sort_by='eq_len_rand_batch',
+                       batch_first=True, shuffle=False, fine_tune=False, keep_in_mem=False,
+                       split_ratio: float = 0., dynamic_epoch=False):
+
+        data_path = self.train_db if self.train_db.exists() else self.train_file
         if fine_tune:
             if not self.finetune_file.exists():
                 # user may have added fine tune file later
                 self._pre_process_parallel('finetune_src', 'finetune_tgt', self.finetune_file)
             log.info("Using Fine tuning corpus instead of training corpus")
-            inp_file = self.finetune_file
-        inp_file = IO.maybe_tmpfs(inp_file)
-        train_data = BatchIterable(inp_file, batch_size=batch_size, sort_by=sort_by,
-                                   batch_first=batch_first, shuffle=shuffle, field=self.tgt_vocab,
-                                   keep_in_mem=keep_in_mem, **self._get_batch_args())
-        if steps > 0:
-            train_data = LoopingIterable(train_data, steps)
+            data_path = self.finetune_file
+
+        if split_ratio > 0:
+            data_path = IO.maybe_tmpfs(data_path)
+            train_file = data_path.with_suffix('.db.tmp')
+            file_creator = partial(self.file_creator, train_file=train_file, split_ratio=split_ratio)
+            train_data = GenerativeBatchIterable(
+                file_creator=file_creator, batches=steps, batch_size=batch_size, field=self.tgt_vocab,
+                dynamic_epoch=dynamic_epoch, batch_first=batch_first, shuffle=shuffle, sort_by=sort_by,
+                **self._get_batch_args())
+        else:
+            data = BatchIterable(
+                data_path=data_path, batch_size=batch_size, field=self.tgt_vocab, sort_by=sort_by,
+                batch_first=batch_first, shuffle=shuffle, **self._get_batch_args())
+            train_data = LoopingIterable(data, steps)
+
         return train_data
 
-    def get_val_data(self, batch_size: int, sort_desc=False, batch_first=True,
+
+    def file_creator(self, train_file, split_ratio, *args, **kwargs):
+        self._pre_process_parallel(*args, src_key='train_src', tgt_key='train_tgt',
+                                   out_file=train_file, split_ratio=split_ratio, **kwargs)
+        return train_file
+
+    def get_val_data(self, batch_size: Union[int, Tuple[int,int]], sort_desc=False, batch_first=True,
                      shuffle=False):
         raw_path = None
         prep = self.config.get('prep', {})
@@ -754,13 +838,18 @@ class TranslationExperiment(BaseExperiment):
             # user may have added fine tune file later
             self._pre_process_parallel('combo_src', 'combo_tgt', self.combo_file)
         combo_file = IO.maybe_tmpfs(self.combo_file)
-        data = BatchIterable(combo_file, batch_size=batch_size, sort_desc=sort_desc,
-                             field=self.tgt_vocab, batch_first=batch_first, shuffle=shuffle,
-                             **self._get_batch_args())
+        data = BatchIterable(
+            combo_file, batch_size=batch_size, sort_desc=sort_desc, field=self.tgt_vocab,
+            batch_first=batch_first, shuffle=shuffle, **self._get_batch_args()
+        )
         if steps > 0:
             data = LoopingIterable(data, steps)
         return data
 
+    def reload(self):
+        exp = type(self)(self.work_dir, read_only=self.read_only)
+        self.__dict__ = exp.__dict__
+
     def copy_vocabs(self, other):
         """
         Copies vocabulary files from self to other
@@ -774,7 +863,7 @@ class TranslationExperiment(BaseExperiment):
                                     (self._tgt_field_file, other._tgt_field_file),
                                     (self._shared_field_file, other._shared_field_file)]:
             if source.exists():
-                IO.copy_file(source, destination)
+                IO.copy_file(source.resolve(), destination.resolve())
                 src_txt_file = source.with_name(source.name.replace('.model', '.vocab'))
                 if src_txt_file.exists():
                     dst_txt_file = destination.with_name(
