@@ -5,6 +5,11 @@ from typing import List, Tuple, Type, Dict, Any, Optional, Iterator
 from pathlib import Path
 import math
 from dataclasses import dataclass, field
+import hashlib
+import warnings
+import sys
+import os
+import portalocker
 
 import torch
 from torch import nn as nn
@@ -18,6 +23,10 @@ from rtg.registry import factories, generators
 Hypothesis = Tuple[float, List[int]]
 StrHypothesis = Tuple[float, str]
 
+if not sys.warnoptions:
+    warnings.simplefilter("default") # Change the filter in this process
+    os.environ["PYTHONWARNINGS"] = "default" # Also affect subprocesses
+
 
 def load_models(models: List[Path], exp: Experiment):
     res = []
@@ -197,8 +206,22 @@ class Decoder:
             if not model_paths:
                 # Average
                 model_paths = exp.list_models()[:ensemble]
-            log.info(f"Averaging {len(model_paths)} model states :: {model_paths}")
-            return Decoder.average_states(model_paths)
+            digest = hashlib.md5(";".join(str(p) for p in model_paths).encode('utf-8')).hexdigest()
+            cache_file = exp.model_dir / f'avg_state{len(model_paths)}_{digest}.pkl'
+            lock_file = cache_file.with_suffix('.lock')
+            MAX_TIMEOUT = 12 * 60 * 60  # 12 hours
+            with portalocker.Lock(lock_file, 'w', timeout=MAX_TIMEOUT) as fh:
+                # check if downloaded by  other parallel process
+                if lock_file.exists() and cache_file.exists():
+                    log.info(f"Cache exists: reading from {cache_file}")
+                    state = Decoder._checkpt_to_model_state(cache_file)
+                else:
+                    log.info(f"Averaging {len(model_paths)} model states :: {model_paths}")
+                    state = Decoder.average_states(model_paths)
+                    if len(model_paths) > 1:
+                        log.info(f"Caching the averaged state at {cache_file}")
+                        torch.save(state, str(cache_file))
+            return state
 
     @classmethod
     def combo_new(cls, exp: Experiment, model_paths: List[str], weights: List[float]):
@@ -318,7 +341,7 @@ class Decoder:
         """
         args = dict((k, v) for k, v in args.items() if v is not None)
         if args:
-            log.warn(f"Ignored args: {args} . To remove this message simply remove the args")
+            warnings.warn(f"Ignored args: {args}. To remove this message simply remove the args")
         assert beam_size >= num_hyp
         device = x_seqs.device
         batch_size = x_seqs.size(0)
