@@ -1,18 +1,18 @@
 # Transformer aka "Attention is all you need"
 # Thanks to http://nlp.seas.harvard.edu/2018/04/03/attention.html
-import os
 import copy
 import math
 import time
 import gc
 from abc import ABC
-from typing import Callable, Optional, List, Union
+from typing import Callable, Optional, Union
 import traceback
 
 import torch
 import torch.nn as nn
 import torch.nn.functional as F
 from torch.autograd import Variable
+from torch.cuda.amp import autocast
 from tqdm import tqdm
 
 from rtg import device, log, TranslationExperiment as Experiment
@@ -24,6 +24,10 @@ from rtg.module.criterion import Criterion
 from torch.optim.optimizer import Optimizer
 from dataclasses import dataclass
 from sacrebleu import corpus_bleu
+from rtg.distrib import DistribTorch
+
+
+dtorch = DistribTorch.instance()
 
 
 def clones(module, N):
@@ -377,7 +381,9 @@ def attention(query, key, value, mask=None, dropout=None):
         # Now, if you got this, take a moment to thank http://nlp.seas.harvard.edu/rush.html
         # for devising this concise code. I needed a lot of time to understand how this code works!
         #
-        scores = scores.masked_fill(mask == 0, -1e9)
+        #scores = scores.masked_fill(mask == 0, -1e9)
+        low_val = -2**15 if dtorch.fp16 else -1e9
+        scores = scores.masked_fill(mask == 0, low_val)
     p_attn = F.softmax(scores, dim=-1)  # [BatchSize x Heads x Time=SeqLen x SeqLen ]
     if dropout is not None:
         p_attn = dropout(p_attn)
@@ -495,10 +501,9 @@ class SimpleLossFunction:
         loss = self.criterion(scores, truth).sum() / normalizer
 
         if train_mode:  # don't do this for validation set
-            loss.backward()
+            dtorch.backward(loss)
             if take_step:
-                self.opt.step()
-                self.opt.zero_grad()
+                dtorch.step(self.opt)
         result = loss.item()
         if get_out:
             result = (result, x_probs.argmax(dim=-1))
@@ -542,13 +547,12 @@ class ChunkedLossCompute(SimpleLossFunction):
             loss = self.criterion(chunked_dist, chunked_ys).sum() / normalizer
             total += loss.detach().item()
             if train_mode:
-                loss.backward()
+                dtorch.backward(loss)
         if train_mode:
             out_grad = _y_feats.grad.data
             y_feats.backward(gradient=out_grad)
             if take_step:
-                self.opt.step()
-                self.opt.zero_grad()
+                dtorch.step(optimizer=self.opt)
         if get_out:
             outs = torch.cat(out_chunks, dim=1)
             return total, outs
@@ -556,100 +560,6 @@ class ChunkedLossCompute(SimpleLossFunction):
             return total
 
 
-class MultiGPULossFunction(ChunkedLossCompute):
-
-    def __init__(self, dp_module: nn.DataParallel, criterion: Criterion, opt: Optimizer,
-                 chunk_size: int, devices: List, out_device=None):
-        super().__init__(None, criterion, opt, chunk_size)
-        self.multi_gpu = len(devices) > 1
-        assert self.multi_gpu
-        self.devices = devices
-        self.out_device = out_device if out_device is not None else devices[0]
-        assert isinstance(dp_module, nn.DataParallel)
-        self.dp_module: nn.DataParallel = dp_module
-        self.sct_criteria = nn.parallel.replicate(self.criterion, devices=self.devices)
-
-    def __call__(self, y_feats, y_seqs, normalizer: Union[int, float],
-                 train_mode=True, chunk_size=None, take_step=True, get_out=False):
-
-        batch_dim = 0
-        assert y_feats.shape[batch_dim] == y_seqs.shape[batch_dim]
-
-        # disconnect y_feats nodes from rest of graph
-        _y_feats = y_feats.data.clone().detach()
-        _y_feats.requires_grad = True  # even though detached, we still need grads here
-
-        # naming: sct = Scattered  chk = Chunked
-        # Scatter is horizontal split (i.e. along batch) ; Chunk is vertical split (ie. along time)
-        # Scatter is handled by pytorch's dataparallel utils
-        sct_feats = nn.parallel.scatter(_y_feats, target_gpus=self.devices, dim=batch_dim)
-        sct_ys = nn.parallel.scatter(y_seqs, target_gpus=self.devices, dim=batch_dim)
-        kwargs_tup = [dict(score=self.criterion.input_type) for _ in sct_feats]
-        assert len(sct_feats) == len(sct_ys)
-
-        n_scts = len(sct_feats)  # if the batch is smaller than n_gpus; only use a subset
-
-        sct_criteria = self.sct_criteria[:n_scts]
-        # use generator from data parallel, because the generator params maybe tied to embeddings
-        # TODO: I am not 100% sure if this actually works
-        generator = self.dp_module.module.generator
-        sct_generators = nn.parallel.replicate(generator, devices=self.devices)[:n_scts]
-
-        chunk_size = chunk_size or self.chunk_size
-        assert chunk_size > 0
-        seq_len = y_feats.shape[1]  # B x L x D
-        assert seq_len == y_seqs.shape[-1]  # B x L
-        total_loss = 0
-        if get_out:
-            sct_chk_outs = [[] for _ in range(n_scts)]   # [S x [T2 x [B2 x C]]
-        # B2 is B / S ... ie. data parallel divided mini-mini batches
-        # T2 is T / C  ... ie  T chunked into mini-T
-        for i in range(0, seq_len, chunk_size):
-            # TODO: remove for loops that runs on CPUs; make it GPU loops
-            # [S x [B2 x C x D]]   ; S is scatter in a python list with items of B x C x D tensors;
-            #                     each scattered is on a different GPU mem
-            chk_sct_feats = [sf[:, i:i + chunk_size] for sf in sct_feats]
-            # [S x [ B2 x C ]] -> [S x [B2.D]]
-            chk_sct_flat_ys = [sy[:, i:i + chunk_size].contiguous().view(-1) for sy in sct_ys]
-            # [S x [B2 x C x D]] --> [S x [B2 x C x V]]
-            chk_sct_dist = nn.parallel.parallel_apply(sct_generators, chk_sct_feats,
-                                                      kwargs_tup=kwargs_tup)
-            if get_out:
-                # [S x [B2 x C x V]] --> [S x [B2 x C]]
-                sct_outs = [chk_dist.argmax(dim=-1) for chk_dist in chk_sct_dist]
-                for i, out in enumerate(sct_outs):
-                    sct_chk_outs[i].append(out)
-
-            chk_sct_flt_dist = [chk_dist.contiguous().view(-1, chk_dist.shape[-1]) for chk_dist in
-                                chk_sct_dist]
-            args_pair = list(zip(chk_sct_flt_dist, chk_sct_flat_ys))
-            chk_sct_loss = nn.parallel.parallel_apply(sct_criteria, args_pair)
-
-            # update total loss
-            chk_losses = nn.parallel.gather(chk_sct_loss, target_device=self.out_device)
-            chk_loss = chk_losses.sum() / normalizer
-            total_loss += chk_loss.item()
-
-            if train_mode:
-                chk_loss.backward()  # backward for the chunked part
-
-        # back prop all loss through the rest of the network
-        if train_mode:
-            # back prop the rest of network
-            y_feats.backward(gradient=_y_feats.grad.data)
-            if take_step:
-                self.opt.step()
-                self.opt.zero_grad()
-        if get_out:
-            # [S x [NumChunks x [B/S x ChunkSize]]]
-            # first cat [NumChunks x [B/S x ChunkSize] along time dim to make full seq
-            sct_outs = [torch.cat(chk_outs, dim=1) for  chk_outs  in sct_chk_outs]  # [S x [B x T]]
-            # next cat along the batch, gather from all GPUs to CPU=-1
-            outs = nn.parallel.gather(sct_outs, target_device=-1, dim=0)
-            return total_loss, outs
-        return total_loss
-
-
 class TransformerTrainer(SteppedTrainer):
 
     def __init__(self, exp: Experiment,
@@ -658,26 +568,21 @@ class TransformerTrainer(SteppedTrainer):
                  model_factory=TransformerNMT.make_model,
                  **optim_args):
         super().__init__(exp, model, model_factory=model_factory, optim=optim, **optim_args)
-        generator = self.model.generator
-        self.n_gpus = torch.cuda.device_count()
         trainer_args = self.exp.config.get('trainer', {}).get('init_args', {})
-        chunk_size = trainer_args.get('chunk_size', 10)
+        chunk_size = trainer_args.get('chunk_size', -1)
         self.grad_accum_interval = trainer_args.get('grad_accum', 1)
         assert self.grad_accum_interval > 0
 
-        log.info(f"Going to use {self.n_gpus} GPUs; "
-                 f" Chunk_size={chunk_size} CUDA_VISIBLE_DEVICES="
-                 f"{os.environ.get('CUDA_VISIBLE_DEVICES')}")
-
         if self.n_gpus > 1:  # Multi GPU mode
-            device_ids = list(range(self.n_gpus))
-            log.warning("Multi GPU mode <<this feature is not well tested>>")
-            self.model = nn.DataParallel(self.model, dim=0, device_ids=device_ids)
+            raise Exception(f"Please use: python -m rtg.distrib.launch -G {self.n_gpus} \n "
+                            f" or set single GPU by: export CUDA_VISIBLE_DEVICES=0 ")
 
-            self.loss_func = MultiGPULossFunction(self.model, criterion=self.criterion,
-                                                  opt=self.opt,
-                                                  chunk_size=chunk_size, devices=device_ids)
+        generator = self.core_model.generator
+        if not chunk_size or chunk_size < 1:
+            self.loss_func = SimpleLossFunction(generator=generator, criterion=self.criterion,
+                                                opt=self.opt)
         else:
+            log.info(f"Using Chunked Loss Generator. chunk_size={chunk_size}")
             self.loss_func = ChunkedLossCompute(generator=generator, criterion=self.criterion,
                                                 opt=self.opt, chunk_size=chunk_size)
 
@@ -692,12 +597,14 @@ class TransformerTrainer(SteppedTrainer):
         total_loss = 0.0
         num_batches = 0
         hyps, refs = [], []   # BLEU
-
+        model = self.core_model
+        assert not model.training
         with tqdm(data_iter, total=data_iter.num_batches,
                   unit='batch', dynamic_ncols=True) as data_bar:
             # TODO: BLEU1
             for i, batch in enumerate(data_bar):
-                batch = batch.to(device)
+                if self.n_gpus <= 1:
+                    batch = batch.to(device)
                 num_toks = batch.y_toks
                 x_seqs = batch.x_seqs
                 if do_bleu and not bool(batch.y_raw):
@@ -711,19 +618,19 @@ class TransformerTrainer(SteppedTrainer):
                     x_seqs = x_seqs[:, 1:]
                 else:
                     bos_step = torch.full((len(batch), 1), fill_value=batch.bos_val,
-                                          dtype=torch.long, device=device)
-
+                                          dtype=torch.long, device=batch.y_seqs.device)
 
                 x_mask = (x_seqs != batch.pad_val).unsqueeze(1)
                 y_seqs_with_bos = torch.cat([bos_step, batch.y_seqs], dim=1)
                 y_mask = batch.make_autoreg_mask(y_seqs_with_bos)
-                out = self.model(x_seqs, y_seqs_with_bos, x_mask, y_mask)
+                out = model(x_seqs, y_seqs_with_bos, x_mask, y_mask)
+
                 # [Batch x Time x D]
                 # skip the last time step (the one with EOS as input)
                 out = out[:, :-1, :]
                 # assumption:  y_seqs has EOS, and not BOS
-
-                loss = self.loss_func(out, batch.y_seqs, num_toks, False, get_out=do_bleu)
+                loss = self.loss_func(out, batch.y_seqs, num_toks, train_mode=False,
+                                      get_out=do_bleu)
                 if do_bleu:
                     loss, outs = loss
                     outs = outs.tolist()
@@ -744,7 +651,7 @@ class TransformerTrainer(SteppedTrainer):
         if do_bleu:
             # this is non standard BLEU: greedy(beam=1), tokenized with whatever was used for training
             bleu = corpus_bleu(hyps, [refs], tokenize='none', force=True)
-            log.info(f'\n {bleu}')
+            log.info(f'\n {bleu.format()}')
             data = {f'P{i+1}':p for i, p in enumerate(bleu.precisions)}
             data['bleu']=  bleu.score
             self.tbd.add_scalars('validn_greedytokbleu', data, self.opt.curr_step)
@@ -788,6 +695,8 @@ class TransformerTrainer(SteppedTrainer):
         """
         log_resources = args.pop('log_resources', False)
         log_embedding = args.pop('log_embedding', False)
+        split_ratio = args.pop('split_ratio', 0.)
+        dynamic_epoch = args.pop('dynamic_epoch', False)
         assert log_interval > 0
 
         # Gradient accumulation
@@ -795,7 +704,10 @@ class TransformerTrainer(SteppedTrainer):
         batches = steps * self.grad_accum_interval
         start_batch = self.start_step * self.grad_accum_interval
         check_point = check_point * self.grad_accum_interval
-
+        if isinstance(batch_size, int):
+            max_toks, max_sents = batch_size, float('inf')
+        else:
+            max_toks, max_sents = batch_size
         if args:
             # no extra args. let user know if an extra arg is passed
             raise Exception(f" Found extra args: {args}")
@@ -804,39 +716,44 @@ class TransformerTrainer(SteppedTrainer):
                  f' batch_size={batch_size} toks; sort_by={sort_by};'
                  f' check point size:{check_point}; fine_tune={fine_tune};'
                  f' dec_bos_cut={dec_bos_cut}')
-        """
-        if self.n_gpus > 1:
-            batch_size *= self.n_gpus
-            log.info(f"# GPUs = {self.n_gpus}, batch_size is set to {batch_size}")
-        """
+
+        distr = DistribTorch.instance()
         if batches <= start_batch:
             raise Exception(f'The model was already trained to {self.start_step} steps. '
                             f'Please increase the steps or clear the existing models')
-        train_data = self.exp.get_train_data(batch_size=batch_size, steps=batches - start_batch,
-                                             sort_by=sort_by, batch_first=True, fine_tune=fine_tune,
-                                             keep_in_mem=keep_in_mem)
-        val_data = self.exp.get_val_data(batch_size, shuffle=False, batch_first=True,
-                                         sort_desc=False)
+
+        train_data = self.exp.get_train_data(
+            batch_size=batch_size, steps=batches - start_batch, sort_by=sort_by, batch_first=True,
+            fine_tune=fine_tune, keep_in_mem=keep_in_mem, split_ratio=split_ratio, dynamic_epoch=dynamic_epoch
+        )
+        val_data = None
+        if distr.is_global_main:
+            val_data = self.exp.get_val_data(batch_size=max_toks, shuffle=False, batch_first=True,
+                                             sort_desc=False)
 
         train_state = TrainerState(self.model, check_point=check_point)
         train_state.train_mode(True)
         unsaved_state = False
         cuda_available = torch.cuda.is_available()
-        update_interval = 0
 
+        batch_count = -1
         stopper = None
         early_stopped = False   # or converged
         if early_stop:
             stopper = EarlyStopper(cur_step=self.start_step, **early_stop)
 
         with tqdm(train_data, initial=start_batch, total=batches, unit='batch',
-                  dynamic_ncols=True) as data_bar:
+                  dynamic_ncols=True, disable=not distr.is_global_main) as data_bar:
             for batch in data_bar:
-                if update_interval == 0:
-                    self.model.zero_grad()
+                batch_count += 1
+                take_step = (batch_count % self.grad_accum_interval) == 0
 
-                # Prep batch
-                batch = batch.to(device)
+               # if update_interval == 0:
+               #     self.model.zero_grad()
+
+                #  if not dataparallel, then move
+                if self.n_gpus <= 1:
+                    batch = batch.to(device)
                 num_toks = batch.y_toks
                 x_seqs = batch.x_seqs
                 if dec_bos_cut:
@@ -844,25 +761,24 @@ class TransformerTrainer(SteppedTrainer):
                     x_seqs = x_seqs[:, 1:]
                 else:
                     bos_step = torch.full((len(batch), 1), fill_value=batch.bos_val,
-                                          dtype=torch.long, device=device)
+                                          dtype=torch.long, device=batch.y_seqs.device)
 
                 # Prep masks
                 x_mask = (x_seqs != batch.pad_val).unsqueeze(1)
                 y_seqs_with_bos = torch.cat([bos_step, batch.y_seqs], dim=1)
                 y_mask = batch.make_autoreg_mask(y_seqs_with_bos)
 
-                # [Batch x Time x D]
-                out = self.model(x_seqs, y_seqs_with_bos, x_mask, y_mask)
+                with autocast(enabled=dtorch.fp16):
+                    # [Batch x Time x D]
+                    out = self.model(x_seqs, y_seqs_with_bos, x_mask, y_mask)
 
-                # skip the last time step (the one with EOS as input)
-                out = out[:, :-1, :]
+                    # skip the last time step (the one with EOS as input)
+                    out = out[:, :-1, :]
 
-                # Trigger optimizer step after gradient accumulation
-                take_step = update_interval == (self.grad_accum_interval - 1)
+                    # assumption:  y_seqs has EOS, and not BOS
+                    loss = self.loss_func(out, batch.y_seqs, num_toks, train_mode=True,
+                                          take_step=take_step)
 
-                # assumption:  y_seqs has EOS, and not BOS
-                loss = self.loss_func(out, batch.y_seqs, num_toks, train_mode=True,
-                                      take_step=take_step)
                 if stopper and take_step:
                     stopper.step()
                 # Log
@@ -875,42 +791,45 @@ class TransformerTrainer(SteppedTrainer):
                         self._log_resources(batch)
 
                 progress_msg, is_check_pt = train_state.step(num_toks, loss)
-                progress_msg += f', LR={self.opt.curr_lr:g}'
+                progress_msg += f', LR={self.opt.curr_lr:0.8f}'
                 data_bar.set_postfix_str(progress_msg, refresh=False)
                 del batch
 
                 # Save checkpoint
                 if is_check_pt:
                     train_loss = train_state.reset()
-                    train_state.train_mode(False)
-                    val_loss = self.run_valid_epoch(val_data, dec_bos_cut=dec_bos_cut)
-                    self.make_check_point(train_loss, val_loss=val_loss, keep_models=keep_models,
-                                          log_embedding=log_embedding)
-                    if check_pt_callback:
-                        check_pt_callback(model=self.model,
-                                          step=self.opt.curr_step,
-                                          train_loss=train_loss)
-                    train_state.train_mode(True)
+                    log.info(f"Chkpt Train loss={train_loss}; Runs validation? {distr.is_global_main}")
+                    if distr.is_global_main:
+                        train_state.train_mode(False)
+                        with torch.no_grad():
+                            val_loss = self.run_valid_epoch(val_data, dec_bos_cut=dec_bos_cut)
+                            self.make_check_point(train_loss, val_loss=val_loss, keep_models=keep_models,
+                                                  log_embedding=log_embedding)
+                            if check_pt_callback:
+                                check_pt_callback(model=self.model,
+                                                  step=self.opt.curr_step,
+                                                  train_loss=train_loss)
+                        train_state.train_mode(True)
+
+                        if stopper:
+                            stopper.validation(val_loss)
+                            if stopper.is_stop():
+                                log.info(f"Stopping at {stopper.cur_step} because {stopper.by}"
+                                         f" didnt improve over {stopper.patience} checkpoints")
+                                early_stopped = True
+                                break
                     unsaved_state = False
                     gc.collect()
-
-                    if stopper:
-                        stopper.validation(val_loss)
-                        if stopper.is_stop():
-                            log.info(f"Stopping at {stopper.cur_step} because {stopper.by}"
-                                     f" didnt improve over {stopper.patience} checkpoints")
-                            early_stopped = True
-                            break
-                # Track gradient accumulation updates
-                update_interval = (update_interval + 1 ) % self.grad_accum_interval
+                    distr.barrier()
 
         # End of training
-        if unsaved_state:
+        if unsaved_state and distr.is_global_main:
             train_loss = train_state.reset()
             train_state.train_mode(False)
             val_loss = self.run_valid_epoch(val_data, dec_bos_cut=dec_bos_cut)
             self.make_check_point(train_loss, val_loss=val_loss, keep_models=keep_models)
 
+        distr.barrier()
         return early_stopped
 
     def _log_resources(self, batch):
